const hid = @import("hid");

const ona = @import("ona");

const ext = @cImport({
	@cInclude("SDL2/SDL.h");
});

const rendering = @import("./rendering.zig");

const std = @import("std");

pub const Assets = struct {
	window: *ext.SDL_Window,
	frame_rendered: std.Thread.ResetEvent = .{},
	pending_work: WorkQueue = .{},
	has_worker_thread: ?std.Thread = null,

	pub const LoadError = std.mem.Allocator.Error;

	pub const LoadFileError = LoadError || ona.files.ReadAllError || error {
		FormatUnsupported,
	};

	pub const TextureFormat = struct {
		extension: []const u8,
		load_file: *const fn (*std.heap.ArenaAllocator, ona.files.Storage, []const u8) LoadFileError!Texture.Desc,
	};

	pub const WorkQueue = ona.asyncio.BlockingQueue(1024, Work);

	fn deinit(self: *Assets) void {
		self.pending_work.enqueue(.shutdown);

		if (self.has_worker_thread) |worker_thread| {
			worker_thread.join();
		}

		self.* = undefined;
	}

	fn init(width: u16, height: u16) !Assets {
		const window = create: {
			const position = ext.SDL_WINDOWPOS_CENTERED;
			const flags = ext.SDL_WINDOW_OPENGL;

			break: create ext.SDL_CreateWindow("Ona", position, position, width, height, flags) orelse {
				return error.Unsupported;
			};
		};

		errdefer {
			ext.SDL_DestroyWindow(window);
		}

		return .{
			.window = window,
		};
	}

	pub fn load_effect_file(self: *Assets, storage: ona.files.Storage, path: []const u8) LoadFileError!Effect {
		if (!std.mem.endsWith(u8, path, ".spv")) {
			return error.FormatUnsupported;
		}

		const fragment_file_stat = try storage.stat(path);
		const fragment_spirv_ops = try ona.heap.allocator.alloc(u32, fragment_file_stat.size / @alignOf(u32));

		defer {
			ona.heap.allocator.free(fragment_spirv_ops);
		}

		const bytes_read = try storage.read_all(path, std.mem.sliceAsBytes(fragment_spirv_ops), .{});

		std.debug.assert(bytes_read.len == fragment_file_stat.size);

		var loaded = ona.asyncio.Future(std.mem.Allocator.Error!Effect){};

		self.pending_work.enqueue(.{
			.load_effect = .{
				.desc = .{
					.fragment_spirv_ops = fragment_spirv_ops,
				},

				.loaded = &loaded,
			},
		});

		return loaded.get().*;
	}

	pub fn load_texture(self: *Assets, desc: Texture.Desc) std.mem.Allocator.Error!Texture {
		var loaded = ona.asyncio.Future(std.mem.Allocator.Error!Texture){};

		self.pending_work.enqueue(.{
			.load_texture = .{
				.desc = desc,
				.loaded = &loaded,
			},
		});

		return loaded.get().*;
	}

	pub const thread_restriction = .main;

	pub fn unload_effect(self: *Assets, handle: Effect) void {
		self.pending_work.enqueue(.{.unload_effect = handle});
	}

	pub fn unload_texture(self: *Assets, handle: Texture) void {
		self.pending_work.enqueue(.{.unload_texture = handle});
	}
};

pub const Color = @Vector(4, f32);

pub const Command = union (enum) {
	draw_font: DrawFont,
	draw_texture: DrawTexture,
	set_effect: SetEffect,
	set_scissor_rect: ?Rect,
	set_target: SetTarget,

	pub const DrawFont = struct {
		position: Vector2 = @splat(0),
		size: Vector2,
		text: []const u8,
		font: Font,
		tint: Color,
		depth: f32,
	};

	pub const DrawTexture = struct {
		anchor: Vector2 = @splat(0),
		size: ?Vector2 = null,
		source: Rect = .{.left = 0, .top = 0, .right = 1, .bottom = 1},
		position: Vector2 = @splat(0),
		tint: Color = colors.white,
		texture: Texture = .default,
		rotation: f32 = 0,
		depth: f32 = 0,
	};

	pub const SetEffect = struct {
		effect: Effect,
		properties: []const u8,
	};

	pub const SetTarget = struct {
		texture: ?Texture = null,
		clear_color: ?Color,
		clear_depth: ?f32,
		clear_stencil: ?u8,
	};
};

pub const Commands = struct {
	pending: *List,

	pub const List = struct {
		arena: std.heap.ArenaAllocator,
		stack: ona.stack.Sequential(Command),

		fn clear(self: *List) void {
			self.stack.clear();

			if (!self.arena.reset(.retain_capacity)) {
				std.log.warn("failed to reset the buffer of a gfx queue with retained capacity", .{});
			}
		}

		fn deinit(self: *List) void {
			self.arena.deinit();
			self.stack.deinit();

			self.* = undefined;
		}

		fn init(allocator: std.mem.Allocator) List {
			return .{
				.arena = std.heap.ArenaAllocator.init(allocator),
				.stack = .{.allocator = allocator},
			};
		}
	};

	pub const Param = struct {
		swap_lists: [2]List,
		swap_state: u1 = 0,

		fn deinit(self: *Param) void {
			for (&self.swap_lists) |*list| {
				list.deinit();
			}

			self.* = undefined;
		}

		fn pending_list(self: *Param) *List {
			return &self.swap_lists[self.swap_state];
		}

		fn rotate(self: *Param) void {
			const swapped_state = self.swap_state ^ 1;

			self.swap_lists[swapped_state].clear();

			self.swap_state = swapped_state;
		}

		pub fn submitted_commands(self: Param) []const Command {
			return self.swap_lists[self.swap_state ^ 1].stack.values;
		}
	};

	const Self = @This();

	pub fn bind(_: ona.World.BindContext) std.mem.Allocator.Error!Param {
		return .{
			.swap_lists = .{
				List.init(ona.heap.allocator),
				List.init(ona.heap.allocator),
			},
		};
	}

	pub fn init(param: *Param) Self {
		return .{
			.pending = param.pending_list(),
		};
	}

	pub fn draw_font(self: Self, command: Command.DrawFont) std.mem.Allocator.Error!void {
		try self.pending.stack.push_grow(.{
			.draw_font = .{
				.text = try self.pending.arena.allocator().dupe(u8, command.text),
				.position = command.position,
				.size = command.size,
				.font = command.font,
				.tint = command.tint,
				.depth = command.depth,
			},
		});
	}

	pub fn draw_texture(self: Self, command: Command.DrawTexture) std.mem.Allocator.Error!void {
		try self.pending.stack.push_grow(.{.draw_texture = command});
	}

	pub fn set_effect(self: Self, command: Command.SetEffect) std.mem.Allocator.Error!void {
		try self.pending.stack.push_grow(.{
			.set_effect = .{
				.properties = try self.pending.arena.allocator().dupe(u8, command.properties),
				.effect = command.effect,
			},
		});
	}

	pub fn set_scissor_rect(self: Self, has_rect: Rect) std.mem.Allocator.Error!void {
		try self.pending.stack.push_grow(.{.set_scissor_rect = has_rect});
	}

	pub fn set_target(self: Self, command: Command.SetTarget) std.mem.Allocator.Error!void {
		try self.pending.stack.push_grow(.{.set_target = command});
	}

	pub fn unbind(param: *Param, _: ona.World.UnbindContext) void {
		param.deinit();
	}
};

pub const Descs = struct {
	arena: std.heap.ArenaAllocator,

	pub const CheckerTextureDesc = struct {
		width: u16,
		height: u16,
		colors: [2]Color,
	};

	pub const BmpTextureDesc = struct {
		storage: ona.files.Storage,
		path: []const u8,
	};

	pub const SolidTextureDesc = struct {
		width: u16,
		height: u16,
		color: Color,
	};

	pub fn bmp_texture(self: *Descs, desc: BmpTextureDesc) (error { OutOfMemory, FormatUnsupported })!Texture.Desc {
		const header = try desc.storage.read_little(desc.path, 0, extern struct {
			type: [2]u8 align (1),
			file_size: u32 align (1),
			reserved: [2]u16 align (1),
			image_offset: u32 align (1),
			header_size: u32 align (1),
			pixel_width: i32 align (1),
			pixel_height: i32 align (1),
			color_planes: u16 align (1),
			bits_per_pixel: u16 align (1),
			compression_method: u32 align (1),
			image_size: u32 align(1),
			pixels_per_meter_x: i32 align (1),
			pixels_per_meter_y: i32 align (1),
			palette_colors_used: u32 align (1),
			important_colors_used: u32 align (1),
		}) orelse {
			return error.FormatUnsupported;
		};

		if (!std.mem.eql(u8, &header.type, "BM")) {
			return error.FormatUnsupported;
		}

		const pixel_width = std.math.cast(u16, header.pixel_width) orelse {
			return error.FormatUnsupported;
		};

		const pixels = try self.arena.allocator().alloc(u8, header.image_size);
		const bytes_per_pixel = header.bits_per_pixel / @bitSizeOf(u8);
		const alignment = 4;
		const byte_stride = pixel_width * bytes_per_pixel;
		const padded_byte_stride = alignment * @divTrunc((byte_stride + alignment - 1), alignment);
		const byte_padding = ona.scalars.sub(padded_byte_stride, byte_stride) orelse 0;
		var buffer_offset: usize = 0;
		var file_offset = @as(usize, header.image_offset);

		switch (header.bits_per_pixel) {
			32 => {
				while (buffer_offset < pixels.len) {
					const line = pixels[buffer_offset .. buffer_offset + byte_stride];

					if (try desc.storage.read(desc.path, line, file_offset) != byte_stride) {
						return error.FormatUnsupported;
					}

					for (0 .. pixel_width) |i| {
						const line_offset = i * 4;
						const pixel = line[line_offset .. line_offset + 4];

						std.mem.swap(u8, &pixel[0], &pixel[2]);
					}

					file_offset += line.len + byte_padding;
					buffer_offset += padded_byte_stride;
				}
			},

			else => return error.FormatUnsupported,
		}

		return .{
			.format = .rgba8,

			.access = .{
				.static = .{
					.width = pixel_width,
					.data = pixels,
				},
			},
		};
	}

	pub fn checker_texture(self: *Descs, desc: CheckerTextureDesc) std.mem.Allocator.Error!Texture.Desc {
		const color_data = try self.arena.allocator().alloc(u32, desc.width * desc.height);

		for (color_data, 0 .. color_data.len) |*color, i| {
			const row = i / desc.width;
    		const col = i % desc.width;

			color.* = colors.compress(desc.colors[(col + (row % 2)) % desc.colors.len]);
		}

		return .{
			.access = .{
				.static = .{
					.data = std.mem.sliceAsBytes(color_data),
					.width = desc.width,
				},
			},

			.format = .rgba8,
		};
	}

	pub fn deinit(self: *Descs) void {
		self.arena.deinit();
	}

	pub fn init(allocator: std.mem.Allocator) Descs {
		return .{
			.arena = std.heap.ArenaAllocator.init(allocator),
		};
	}

	pub fn solid_texture(self: *Descs, desc: SolidTextureDesc) std.mem.Allocator.Error!Texture.Desc {
		const color_data = try self.arena.allocator().alloc(u32, desc.width * desc.height);

		@memset(color_data, colors.compress(desc.color));

		return .{
			.access = .{
				.static = .{
					.data = std.mem.sliceAsBytes(color_data),
					.width = desc.width,
				},
			},

			.format = .rgba8,
		};
	}
};

pub const Display = struct {
	width: u16 = 1280,
	height: u16 = 720,
	clear_color: Color = colors.black,
};

pub const Effect = enum (u32) {
	default,
	_,

	pub const Desc = struct {
		fragment_spirv_ops: []const u32,
	};
};

pub const Font = enum (u32) {
	default,
	_,

	pub const Desc = struct {

	};
};

pub const Rect = extern struct {
	left: f32,
	top: f32,
	right: f32,
	bottom: f32,

	pub fn area(self: Rect) f32 {
		const width, const height = self.size();

		return width * height;
	}

	pub fn size(self: Rect) Vector2 {
		return .{self.right - self.left, self.bottom - self.top};
	}

	pub const zero = Rect{
		.left = 0,
		.top = 0,
		.right = 0,
		.bottom = 0,
	};
};

pub const Texture = enum (u32) {
	default,
	_,

	pub const Desc = struct {
		format: Format,
		access: Access,

		pub const Access = union (enum) {
			static: StaticAccess,
			render: RenderAccess,
		};

		pub const StaticAccess = struct {
			width: u16,
			data: []const u8,
		};

		pub const RenderAccess = struct {
			width: u16,
			height: u16,
		};
	};

	pub const Format = enum {
		rgba8,
		bgra8,

		pub fn byte_size(self: Format) usize {
			return switch (self) {
				.rgba8, .bgra8 => 4,
			};
		}
	};
};

pub const Vector2 = @Vector(2, f32);

const Work = union (enum) {
	load_effect: LoadEffect,
	load_texture: LoadTexture,
	render_frame: RenderFrame,
	shutdown,
	unload_effect: UnloadEffect,
	unload_texture: UnloadTexture,

	const LoadEffect = struct {
		desc: Effect.Desc,
		loaded: *ona.asyncio.Future(std.mem.Allocator.Error!Effect),
	};

	const LoadTexture = struct {
		desc: Texture.Desc,
		loaded: *ona.asyncio.Future(std.mem.Allocator.Error!Texture),
	};

	const RenderFrame = struct {
		clear_color: Color,
		width: u16,
		height: u16,
		finished: *std.Thread.ResetEvent,
		has_commands: ?*ona.Params(Commands).Node,
	};

	const UnloadEffect = struct {
		handle: Effect,
	};

	const UnloadTexture = struct {
		handle: Texture,
	};
};

pub const colors = struct {
	pub const black = greyscale(0);

	pub fn compress(color: Color) u32 {
		const range: Color = @splat(255);
		const r, const g, const b, const a = color * range;

		return @bitCast([_]u8{@intFromFloat(r), @intFromFloat(g), @intFromFloat(b), @intFromFloat(a)});
	}

	pub const grey = greyscale(0.5);

	pub fn greyscale(v: f32) Color {
		return .{v, v, v, 1};
	}

	pub const purple = rgb(0.5, 0, 0.5);

	pub const red = rgb(1, 0, 0);

	pub fn rgb(r: f32, g: f32, b: f32) Color {
		return .{r, g, b, 1};
	}

	pub const white = greyscale(1);
};

pub fn poll(loop: ona.Write(ona.Loop), events: ona.Send(hid.Event)) !void {
	var event = @as(ext.SDL_Event, undefined);

	while (ext.SDL_PollEvent(&event) != 0) {
		switch (event.type) {
			ext.SDL_QUIT => {
				loop.state.quit();
			},

			ext.SDL_KEYUP => {
				try events.push(.{.key_up = @enumFromInt(event.key.keysym.scancode)});
			},

			ext.SDL_KEYDOWN => {
				try events.push(.{.key_down = @enumFromInt(event.key.keysym.scancode)});
			},

			ext.SDL_MOUSEBUTTONUP => {
				try events.push(.{
					.mouse_up = switch (event.button.button) {
						ext.SDL_BUTTON_LEFT => .left,
						ext.SDL_BUTTON_RIGHT => .right,
						ext.SDL_BUTTON_MIDDLE => .middle,
						else => unreachable,
					},
				});
			},

			ext.SDL_MOUSEBUTTONDOWN => {
				try events.push(.{
					.mouse_down = switch (event.button.button) {
						ext.SDL_BUTTON_LEFT => .left,
						ext.SDL_BUTTON_RIGHT => .right,
						ext.SDL_BUTTON_MIDDLE => .middle,
						else => unreachable,
					},
				});
			},

			ext.SDL_MOUSEMOTION => {
				try events.push(.{
					.mouse_motion = .{
						.relative_position = .{@floatFromInt(event.motion.xrel), @floatFromInt(event.motion.yrel)},
						.absolute_position = .{@floatFromInt(event.motion.x), @floatFromInt(event.motion.y)},
					},
				});
			},

			else => {},
		}
	}
}

pub fn setup(world: *ona.World) !void {
	if (ext.SDL_Init(ext.SDL_INIT_VIDEO | ext.SDL_INIT_EVENTS) != 0) {
		return error.Unsupported;
	}

	const display = try world.set_get_state(Display{});

	const assets = create: {
		var assets = try Assets.init(display.width, display.height);

		errdefer {
			assets.deinit();
		}

		break: create try world.set_get_state(assets);
	};

	assets.frame_rendered.set();

	errdefer {
		assets.deinit();
	}

	assets.has_worker_thread = try std.Thread.spawn(.{}, rendering.process_work, .{
		&assets.pending_work,
		assets.window,
	});

	try world.on_event(.pre_update, ona.system_fn(poll), .{.label = "poll gfx"});
	try world.on_event(.exit, ona.system_fn(stop), .{.label = "stop gfx"});
	try world.on_event(.finish, ona.system_fn(synchronize), .{.label = "synchronize gfx"});
}

pub fn stop(assets: ona.Write(Assets)) void {
	assets.state.deinit();
}

pub fn synchronize(exclusive: ona.Exclusive(&.{Assets, Display})) !void {
	const assets, const display = exclusive.states;

	assets.frame_rendered.wait();
	assets.frame_rendered.reset();

	var has_command_param = exclusive.world.get_params(Commands).has_head;

	while (has_command_param) |command_param| : (has_command_param = command_param.has_next) {
		command_param.param.rotate();
	}

	var display_width, var display_height = [_]c_int{0, 0};

	ext.SDL_GL_GetDrawableSize(assets.window, &display_width, &display_height);

	if (display.width != display_width or display.height != display_height) {
		ext.SDL_SetWindowSize(assets.window, display.width, display.height);
	}

	if (exclusive.world.get_params(Commands).has_head) |has_commands| {
		assets.pending_work.enqueue(.{
			.render_frame = .{
				.has_commands = has_commands,
				.width = display.width,
				.height = display.height,
				.clear_color = display.clear_color,
				.finished = &assets.frame_rendered,
			},
		});
	} else {
		assets.frame_rendered.set();
	}
}
