const gfx = @import("gfx");

const ona = @import("ona");

const std = @import("std");

pub const Box = struct {
	x: f32,
	y: f32,
	width: f32,
	height: f32,

	pub fn rect(self: Box) gfx.Rect {
		return .{
			.left = self.x,
			.top = self.y,
			.right = self.x + self.width,
			.bottom = self.y + self.height,
		};
	}
};

pub const Canvas = struct {
	nodes: NodeMap = NodeMap.init(ona.heap.allocator),
	drawable_blocks: DrawableBlockMap = DrawableBlockMap.init(ona.heap.allocator),
	drawable_labels: DrawableLabelMap = DrawableLabelMap.init(ona.heap.allocator),

	fn Drawable(comptime State: type) type {
		return struct {
			clip: gfx.Rect,
			box: Box,
			depth: f32,
			state: State,
		};
	}

	const DrawableBlockMap = ona.SlotMap(Drawable(struct {
		background: gfx.Texture = .default,
		color: gfx.Color = gfx.colors.white,
	}));

	const DrawableLabelMap = ona.SlotMap(Drawable(struct {
		text: ona.stack.Sequential(u8) = .{},
		color: gfx.Color = gfx.colors.black,
		font: gfx.Font = .default,
	}));

	const NodeMap = ona.SlotMap(struct {
		name: []u8,
		box: Box,
		parent_index: u32 = unused_index,
		block_index: u32 = unused_index,
		label_index: u32 = unused_index,

		const Self = @This();

		fn clip(self: Self, nodes: NodeMap) gfx.Rect {
			if (self.parent_index == unused_index) {
				return self.box.rect();
			}

			return nodes.get(self.parent_index).?.box.rect();
		}

		fn depth(self: Self, nodes: NodeMap) f32 {
			var parent_index = self.parent_index;
			var value: f32 = 0;

			while (parent_index != unused_index) : (parent_index = nodes.get(parent_index).?.parent_index) {
				value += 1;
			}

			return value;
		}
	});

	pub fn append(self: *Canvas, name: []const u8, box: Box) std.mem.Allocator.Error!Item {
		const duped_name = try ona.heap.allocator.dupe(u8, name);

		errdefer {
			ona.heap.allocator.free(duped_name);
		}

		return @enumFromInt(try self.nodes.insert(.{
			.name = duped_name,
			.box = box,
		}));
	}

	fn deinit(self: *Canvas) void {
		self.nodes.deinit();
		self.drawable_blocks.deinit();

		for (self.drawable_labels.values()) |*drawable_label| {
			drawable_label.state.text.deinit();
		}

		self.drawable_labels.deinit();

		self.* = undefined;
	}

	pub fn set_block(self: *Canvas, item: Item, set: SetBlock) std.mem.Allocator.Error!void {
		const node = self.nodes.get(@intFromEnum(item)).?;

		if (node.block_index == unused_index) {
			node.block_index = @intCast(try self.drawable_blocks.insert(.{
				.clip = node.clip(self.nodes),
				.depth = node.depth(self.nodes),
				.box = node.box,
				.state = .{},
			}));
		}

		const block = self.drawable_blocks.get(node.block_index).?;

		if (set.has_background) |background| {
			block.state.background = background;
		}

		if (set.has_color) |color| {
			block.state.color = color;
		}
	}

	pub fn set_label(self: *Canvas, item: Item, set: SetLabel) std.mem.Allocator.Error!void {
		const node = self.nodes.get(@intFromEnum(item)).?;

		if (node.label_index == unused_index) {
			node.label_index = @intCast(try self.drawable_labels.insert(.{
				.clip = node.clip(self.nodes),
				.depth = node.depth(self.nodes),
				.box = node.box,
				.state = .{},
			}));
		}

		const label = self.drawable_labels.get(node.label_index).?;

		if (set.has_text) |text| {
			label.state.text.pop_all();

			try label.state.text.grow(ona.scalars.sub(text.len, label.state.text.cap) orelse 0);

			std.debug.assert(label.state.text.push_all(text));
		}

		if (set.has_font) |font| {
			label.state.font = font;
		}

		if (set.has_color) |color| {
			label.state.color = color;
		}
	}

	pub fn set_parent(self: *Canvas, item: Item, has_parent: ?Item) void {
		const node = self.nodes.get(@intFromEnum(item)).?;

		node.parent_index = if (has_parent) |parent| @intFromEnum(parent) else unused_index;

		const depth = node.depth(self.nodes);
		const clip = node.clip(self.nodes);

		if (node.block_index != unused_index) {
			const block = self.drawable_blocks.get(node.block_index).?;

			block.depth = depth;
			block.clip = clip;
		}

		if (node.label_index != unused_index) {
			const label = self.drawable_labels.get(node.label_index).?;

			label.depth = depth;
			label.clip = clip;
		}
	}

	const unused_index = std.math.maxInt(u32);
};

pub const Item = enum (u32) {
	_,
};

pub const SetBlock = struct {
	has_background: ?gfx.Texture = null,
	has_color: ?gfx.Color = null,
};

pub const SetLabel = struct {
	has_text: ?[]const u8 = null,
	has_color: ?gfx.Color = null,
	has_font: ?gfx.Font = null,
};

pub fn exit(canvas: ona.Write(Canvas)) !void {
	canvas.state.deinit();
}

pub fn render(commands: gfx.Commands, canvas: ona.Read(Canvas)) !void {
	// TODO: Investigate if scissor rects are necessary for clipping content that overflows.
	for (canvas.state.drawable_blocks.values()) |block| {
		try commands.set_scissor_rect(block.clip);

		try commands.draw_texture(.{
			.size = .{block.box.width, block.box.height},
			.position = .{block.box.x, block.box.y},
			.depth = block.depth,
			.texture = block.state.background,
			.tint = block.state.color,
		});
	}

	for (canvas.state.drawable_labels.values()) |label| {
		try commands.set_scissor_rect(label.clip);

		try commands.draw_font(.{
			.size = .{label.box.width, label.box.height},
			.depth = label.depth,
			.text = label.state.text.values,
			.position = .{label.box.x, label.box.y},
			.tint = label.state.color,
			.font = label.state.font,
		});
	}
}

pub fn setup(world: *ona.World) !void {
	try world.set_state(Canvas{});
	try world.on_event(.render, ona.system_fn(render), .{.label = "render gui"});
	try world.on_event(.exit, ona.system_fn(exit), .{.label = "cleanup gui"});
}
