const builtin = @import("builtin");

const ona = @import("./ona.zig");

const std = @import("std");

thread_pool: ?*std.Thread.Pool = null,
thread_restricted_resources: [std.enums.values(ona.ThreadRestriction).len]StateTable,
event_schedules: ona.map.Hashed([]const u8, *Schedule, ona.map.string_traits),

pub const BindContext = struct {
	node: ona.dag.Node,
	systems: *Schedule,
	world: *Self,

	pub fn accesses_state(self: BindContext, access: std.meta.Tag(StateAccess), id: ona.TypeID) bool {
		const resource_accesses = &self.systems.graph.get(self.node).?.resource_accesses;

		for (resource_accesses.values) |resource_access| {
			switch (resource_access) {
				.read_only => |resource_id| {
					if (access == .read_only and resource_id == id) {
						return true;
					}
				},

				.read_write => |resource_id| {
					if (access == .read_write and resource_id == id) {
						return true;
					}
				},
			}
		}

		return false;
	}

	pub fn register_writable_state_access(self: BindContext, comptime Resource: type) std.mem.Allocator.Error!?*Resource {
		const value = self.world.get_state(Resource) orelse {
			return null;
		};

		const id = ona.type_id(Resource);

		if (!self.accesses_state(.read_write, id)) {
			try self.systems.graph.get(self.node).?.resource_accesses.push_grow(.{.read_write = id});
		}

		const read_write_resource_nodes = lazily_create: {
			break: lazily_create self.systems.read_write_resource_id_nodes.get(id) orelse insert: {
				std.debug.assert(try self.systems.read_write_resource_id_nodes.emplace(id, .{
					.allocator = ona.heap.allocator,
				}));

				break: insert self.systems.read_write_resource_id_nodes.get(id).?;
			};
		};

		if (std.mem.indexOfScalar(ona.dag.Node, read_write_resource_nodes.values, self.node) == null) {
			try read_write_resource_nodes.push_grow(self.node);
		}

		return value;
	}

	pub fn register_readable_state_access(self: BindContext, comptime Resource: type) std.mem.Allocator.Error!?*const Resource {
		const value = self.world.get_state(Resource) orelse {
			return null;
		};

		const id = ona.type_id(Resource);

		if (!self.accesses_state(.read_only, id)) {
			try self.systems.graph.get(self.node).?.resource_accesses.push_grow(.{.read_only = id});
		}

		const read_only_resource_nodes = lazily_create: {
			break: lazily_create self.systems.read_only_resource_id_nodes.get(id) orelse insert: {
				std.debug.assert(try self.systems.read_only_resource_id_nodes.emplace(id, .{
					.allocator = ona.heap.allocator,
				}));

				break: insert self.systems.read_only_resource_id_nodes.get(id).?;
			};
		};

		if (std.mem.indexOfScalar(ona.dag.Node, read_only_resource_nodes.values, self.node) == null) {
			try read_only_resource_nodes.push_grow(self.node);
		}

		return value;
	}
};

const Dependency = struct {
	kind: Kind,
	id: usize,

	const Kind = enum {
		after,
		before,
	};
};

const Graph = ona.dag.Graph(struct {
	info: *const ona.SystemInfo,
	label: [:0]u8,
	dependencies: []Dependency,
	parameter_states: [ona.SystemInfo.max_parameters]*anyopaque = [_]*anyopaque{undefined} ** ona.SystemInfo.max_parameters,
	resource_accesses: ona.stack.Sequential(StateAccess),
});

const ParallelNodeBundles = ona.stack.Sequential(NodeBundle);

const NodeBundle = ona.stack.Sequential(ona.dag.Node);

const StateAccess = union (enum) {
	read_only: ona.TypeID,
	read_write: ona.TypeID,
};

const ResourceNodeBundle = ona.map.Hashed(ona.TypeID, NodeBundle, ona.map.enum_traits(ona.TypeID));

const Schedule = struct {
	label: [:0]const u8,
	graph: Graph,
	arena: std.heap.ArenaAllocator,
	system_id_nodes: ona.map.Hashed(usize, NodeBundle, ona.map.usize_traits),
	inserted_systems: ona.stack.Sequential(InsertedSystem) = .{},
	read_write_resource_id_nodes: ResourceNodeBundle,
	read_only_resource_id_nodes: ResourceNodeBundle,
	parallel_work_bundles: ParallelNodeBundles,
	blocking_work: NodeBundle,

	const InsertedSystem = struct {
		info: *const ona.SystemInfo,
		order: ona.SystemOrder,
	};

	pub fn deinit(self: *Schedule, world: *Self) void {
		{
			var nodes = self.system_id_nodes.entries();

			while (nodes.next()) |node| {
				node.value.deinit();
			}
		}

		{
			var nodes = self.read_write_resource_id_nodes.entries();

			while (nodes.next()) |node| {
				node.value.deinit();
			}
		}

		{
			var nodes = self.read_only_resource_id_nodes.entries();

			while (nodes.next()) |node| {
				node.value.deinit();
			}
		}

		var nodes = self.graph.nodes();

		while (nodes.next()) |node| {
			const system = self.graph.get(node).?;

			for (system.info.used_parameters(), system.parameter_states[0 .. system.info.parameter_count]) |parameter, state| {
				parameter.unbind(self.arena.allocator(), state, .{
					.world = world,
				});
			}

			system.resource_accesses.deinit();
			ona.heap.allocator.free(system.dependencies);
			ona.heap.allocator.free(system.label);
		}

		for (self.parallel_work_bundles.values) |*bundle| {
			bundle.deinit();
		}

		self.parallel_work_bundles.deinit();
		self.blocking_work.deinit();
		self.graph.deinit();
		self.system_id_nodes.deinit();
		self.read_write_resource_id_nodes.deinit();
		self.read_only_resource_id_nodes.deinit();
		self.inserted_systems.deinit();
		self.arena.deinit();
	}

	pub fn run(self: *Schedule, world: *Self) anyerror!void {
		if (self.is_invalidated()) {
			const work = struct {
				fn bind_systems(schedule: *Schedule, world_: *Self) !void {
					for (schedule.inserted_systems.values) |system| {
						const nodes = lazily_create: {
							const system_id = @intFromPtr(system.info);

							break: lazily_create schedule.system_id_nodes.get(system_id) orelse insert: {
								std.debug.assert(try schedule.system_id_nodes.emplace(system_id, .{
									.allocator = schedule.system_id_nodes.allocator,
								}));

								break: insert schedule.system_id_nodes.get(system_id).?;
							};
						};

						const dependencies = init: {
							const total_run_orders = system.order.run_after.len + system.order.run_before.len;
							const dependencies = try ona.heap.allocator.alloc(Dependency, total_run_orders);
							var dependencies_written = @as(usize, 0);

							for (system.order.run_after) |after_system| {
								dependencies[dependencies_written] = .{
									.id = @intFromPtr(after_system),
									.kind = .after,
								};

								dependencies_written += 1;
							}

							for (system.order.run_before) |before_system| {
								dependencies[dependencies_written] = .{
									.id = @intFromPtr(before_system),
									.kind = .before,
								};

								dependencies_written += 1;
							}

							break: init dependencies;
						};

						errdefer ona.heap.allocator.free(dependencies);

						const label = try ona.heap.allocator.dupeZ(u8, if (system.order.label.len == 0) "anonymous system" else system.order.label);

						errdefer ona.heap.allocator.free(label);

						const node = try schedule.graph.append(.{
							.info = system.info,
							.label = label,
							.dependencies = dependencies,
							.resource_accesses = .{.allocator = ona.heap.allocator},
						});

						const system_node = schedule.graph.get(node).?;
						const system_parameter_states = system_node.parameter_states[0 .. system.info.parameter_count];

						errdefer {
							for (system.info.used_parameters(), system_parameter_states) |parameter, state| {
								parameter.unbind(schedule.arena.allocator(), state, .{
									.world = world_,
								});
							}

							std.debug.assert(schedule.graph.remove_node(node) != null);
						}

						for (system_parameter_states, system.info.used_parameters()) |*state, parameter| {
							state.* = try parameter.bind(schedule.arena.allocator(), .{
								.world = world_,
								.node = node,
								.systems = schedule,
							});
						}

						try nodes.push_grow(node);
					}
				}

				fn regenerate_graph(schedule: *Schedule) !void {
					schedule.graph.clear_edges();

					var nodes = schedule.graph.nodes();

					while (nodes.next()) |node| {
						const system = schedule.graph.get(node).?;

						for (system.dependencies) |order| {
							const dependencies = schedule.system_id_nodes.get(@intFromPtr(system.info)) orelse {
								@panic("unable to resolve missing explicit system dependency");
							};

							if (dependencies.is_empty()) {
								@panic("unable to resolve missing explicit system dependency");
							}

							switch (order.kind) {
								.after => {
									for (dependencies.values) |dependency_node| {
										std.debug.assert(try schedule.graph.insert_edge(node, dependency_node));
									}
								},

								.before => {
									for (dependencies.values) |dependency_node| {
										std.debug.assert(try schedule.graph.insert_edge(dependency_node, node));
									}
								},
							}
						}

						for (system.resource_accesses.values) |resource_access| {
							switch (resource_access) {
								.read_write => |resource_id| {
									const read_write_dependencies = schedule.read_write_resource_id_nodes.get(resource_id) orelse {
										@panic("unable to resolve missing implicit read-write parameter dependency");
									};

									for (read_write_dependencies.values) |dependency_node| {
										std.debug.assert(try schedule.graph.insert_edge(node, dependency_node));
									}

									if (schedule.read_only_resource_id_nodes.get(resource_id)) |dependencies| {
										for (dependencies.values) |dependency_node| {
											std.debug.assert(try schedule.graph.insert_edge(node, dependency_node));
										}
									}
								},

								.read_only => |resource_id| {
									if (schedule.read_only_resource_id_nodes.get(resource_id)) |dependencies| {
										for (dependencies.values) |dependency_node| {
											std.debug.assert(try schedule.graph.insert_edge(node, dependency_node));
										}
									}
								},
							}
						}
					}
				}

				fn populate_bundle(bundle: *NodeBundle, graph: *Graph, node: ona.dag.Node) !void {
					std.debug.assert(graph.mark_visited(node));

					for (graph.edge_nodes(node).?) |edge| {
						if (graph.visited(edge).?) {
							continue;
						}

						try populate_bundle(bundle, graph, edge);
					}

					try bundle.push_grow(node);
				}

				fn sort(schedule: *Schedule) !void {
					defer schedule.graph.reset_visited();

					var nodes = schedule.graph.nodes();

					while (nodes.next()) |node| {
						if (schedule.graph.visited(node).?) {
							continue;
						}

						try schedule.parallel_work_bundles.push_grow(.{.allocator = ona.heap.allocator});

						const bundle = schedule.parallel_work_bundles.get().?;

						errdefer {
							bundle.deinit();
							std.debug.assert(schedule.parallel_work_bundles.pop() != null);
						}

						try populate_bundle(bundle, &schedule.graph, node);
					}

					for (schedule.parallel_work_bundles.values) |*work| {
						var index = @as(usize, 0);

						while (index < work.len()) : (index += 1) {
							const node = work.values[index];

							switch (schedule.graph.get(node).?.info.thread_restriction) {
								.none => continue,

								.main => {
									const extracted_work = work.values[index ..];

									try schedule.blocking_work.grow(extracted_work.len);

									std.debug.assert(schedule.blocking_work.push_all(extracted_work));
									std.debug.assert(work.pop_many(extracted_work.len));
								},
							}
						}
					}
				}
			};

			try work.bind_systems(self, world);
			try work.regenerate_graph(self);
			try work.sort(self);
		}

		// TODO: Refactor so the thread pool is a global resource rather than owned per-world.
		if (world.thread_pool) |thread_pool| {
			const parallel = struct {
				fn run(work_group: *std.Thread.WaitGroup, graph: Graph, bundle: NodeBundle) void {
					defer work_group.finish();

					for (bundle.values) |node| {
						const system = graph.get(node).?;

						// TODO: std lib thread pool sucks for many reasons and this is one of them.
						system.info.execute(system.info.used_parameters(), &system.parameter_states) catch unreachable;
					}
				}
			};

			var work_group = std.Thread.WaitGroup{};

			for (self.parallel_work_bundles.values) |bundle| {
				work_group.start();

				try thread_pool.spawn(parallel.run, .{&work_group, self.graph, bundle});
			}

			thread_pool.waitAndWork(&work_group);
		} else {
			for (self.parallel_work_bundles.values) |bundle| {
				for (bundle.values) |node| {
					const system = self.graph.get(node).?;

					try system.info.execute(system.info.used_parameters(), &system.parameter_states);
				}
			}
		}

		for (self.blocking_work.values) |node| {
			const system = self.graph.get(node).?;

			try system.info.execute(system.info.used_parameters(), &system.parameter_states);
		}
	}

	pub fn init(comptime label: [:0]const u8) std.mem.Allocator.Error!Schedule {
		return .{
			.graph = Graph.init(ona.heap.allocator),
			.arena = std.heap.ArenaAllocator.init(ona.heap.allocator),
			.label = label,
			.system_id_nodes = .{},
			.read_write_resource_id_nodes = .{},
			.read_only_resource_id_nodes = .{},
			.parallel_work_bundles = .{},
			.blocking_work = .{},
		};
	}

	pub fn insert(self: *Schedule, info: *const ona.SystemInfo, order: ona.SystemOrder) std.mem.Allocator.Error!void {
		try self.inserted_systems.push_grow(.{
			.info = info,
			.order = order,
		});

		self.blocking_work.clear();

		for (self.parallel_work_bundles.values) |*bundle| {
			bundle.deinit();
		}

		self.parallel_work_bundles.clear();
	}

	pub fn is_invalidated(self: Schedule) bool {
		return self.parallel_work_bundles.is_empty() and self.blocking_work.is_empty();
	}
};

const Self = @This();

const StateTable = struct {
	arena: std.heap.ArenaAllocator,
	table: ona.map.Hashed(ona.TypeID, Entry, ona.map.enum_traits(ona.TypeID)),

	const Entry = struct {
		ptr: *anyopaque,
	};

	fn deinit(self: *StateTable) void {
		self.table.deinit();
		self.arena.deinit();

		self.* = undefined;
	}

	fn get(self: StateTable, comptime Resource: type) ?*Resource {
		if (self.table.get(ona.type_id(Resource))) |entry| {
			return @ptrCast(@alignCast(entry.ptr));
		}

		return null;
	}

	fn init() StateTable {
		return .{
			.arena = std.heap.ArenaAllocator.init(ona.heap.allocator),
			.table = .{.allocator = ona.heap.allocator},
		};
	}

	fn set_get(self: *StateTable, value: anytype) std.mem.Allocator.Error!*@TypeOf(value) {
		try self.set(value);

		return self.get(@TypeOf(value)).?;
	}

	fn set(self: *StateTable, value: anytype) std.mem.Allocator.Error!void {
		const Value = @TypeOf(value);
		const value_id = ona.type_id(Value);

		if (self.table.get(value_id)) |entry| {
			@as(*Value, @ptrCast(@alignCast(entry.ptr))).* = value;
		} else {
			const resource_allocator = self.arena.allocator();
			const allocated_resource = try resource_allocator.create(Value);

			errdefer resource_allocator.destroy(allocated_resource);

			std.debug.assert(try self.table.emplace(value_id, .{
				.ptr = allocated_resource,
			}));

			allocated_resource.* = value;
		}
	}
};

pub const UnbindContext = struct {
	world: *Self,
};

pub fn deinit(self: *Self) void {
	var event_schedules = self.event_schedules.entries();

	while (event_schedules.next()) |schedule| {
		schedule.value.deinit(self);
		ona.heap.allocator.destroy(schedule.value);
	}

	for (&self.thread_restricted_resources) |*resources| {
		resources.deinit();
	}

	if (self.thread_pool) |thread_pool| {
		thread_pool.deinit();
		ona.heap.allocator.destroy(thread_pool);
	}

	self.event_schedules.deinit();

	self.* = undefined;
}

pub fn get_params(self: Self, comptime Value: type) ona.Params(Value) {
	const params = self.get_state(ona.Params(Value)) orelse {
		return .{};
	};

	return params.*;
}

pub fn get_state(self: Self, comptime Value: type) ?*Value {
	return @ptrCast(@alignCast(self.thread_restricted_resources[@intFromEnum(ona.thread_restriction(Value))].get(Value)));
}

pub fn set_get_state(self: *Self, value: anytype) std.mem.Allocator.Error!*@TypeOf(value) {
	return self.thread_restricted_resources[@intFromEnum(ona.thread_restriction(@TypeOf(value)))].set_get(value);
}

pub fn init(thread_count: u32) std.Thread.SpawnError!Self {
	var world = Self{
		.thread_restricted_resources = .{StateTable.init(), StateTable.init()},
		.event_schedules = .{.allocator = ona.heap.allocator},
	};

	if (thread_count != 0 and !builtin.single_threaded) {
		const thread_pool = try ona.heap.allocator.create(std.Thread.Pool);

		try thread_pool.init(.{
			.allocator = ona.heap.allocator,
			.n_jobs = thread_count,
		});

		world.thread_pool = thread_pool;
	}

	return world;
}

pub fn on_event(self: *Self, comptime event: anytype, info: *const ona.SystemInfo, order: ona.SystemOrder) std.mem.Allocator.Error!void {
	const event_name = @tagName(event);

	if (self.event_schedules.get(event_name)) |schedule| {
		try schedule.*.insert(info, order);
	} else {
		const schedule = try ona.heap.allocator.create(Schedule);

		errdefer {
			ona.heap.allocator.destroy(schedule);
		}

		schedule.* = try Schedule.init(event_name);

		errdefer {
			schedule.deinit(self);
		}

		try schedule.insert(info, order);

		std.debug.assert(try self.event_schedules.emplace(event_name, schedule));
	}
}

pub fn run_event(self: *Self, comptime event: anytype) anyerror!void {
	if (self.event_schedules.get(@tagName(event))) |schedule| {
		try schedule.*.run(self);
	}
}

pub fn set_state(self: *Self, value: anytype) std.mem.Allocator.Error!void {
	try self.thread_restricted_resources[@intFromEnum(ona.thread_restriction(@TypeOf(value)))].set(value);
}
